@@ -1,6 +1,6 @@
 /*
     SDL_mixer:  An audio mixer library based on the SDL library
-    Copyright (C) 1997, 1998, 1999, 2000, 2001  Sam Lantinga
+    Copyright (C) 1997-2004 Sam Lantinga
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Library General Public
@@ -20,7 +20,7 @@
     slouken@libsdl.org
 */
 
-/* $Id: music.c,v 1.38 2001/12/17 02:39:12 slouken Exp $ */
+/* $Id: music.c,v 1.58 2004/11/15 23:34:56 slouken Exp $ */
 
 #include <stdlib.h>
 #include <string.h>
@@ -31,6 +31,14 @@
 
 #include "SDL_mixer.h"
 
+#define SDL_SURROUND
+
+#ifdef SDL_SURROUND
+#define MAX_OUTPUT_CHANNELS 6
+#else
+#define MAX_OUTPUT_CHANNELS 2
+#endif
+
 /* The music command hack is UNIX specific */
 #ifndef unix
 #undef CMD_MUSIC
@@ -42,12 +50,17 @@
 #ifdef WAV_MUSIC
 #include "wavestream.h"
 #endif
-#ifdef MOD_MUSIC
+#if defined(MOD_MUSIC) || defined(LIBMIKMOD_MUSIC)
 #  include "mikmod.h"
 #  if defined(LIBMIKMOD_VERSION)                /* libmikmod 3.1.8 */
 #    define UNIMOD			MODULE
 #    define MikMod_Init()		MikMod_Init(NULL)
 #    define MikMod_LoadSong(a,b)	Player_Load(a,b,0)
+#    ifdef USE_RWOPS
+#      ifndef LIBMIKMOD_MUSIC
+#      define MikMod_LoadSongRW(a,b)	Player_LoadRW(a,b,0)
+#    endif
+#    endif
 #    define MikMod_FreeSong		Player_Free
      extern int MikMod_errno;
 #  else                                        /* old MikMod 3.0.3 */
@@ -87,14 +100,7 @@ static int music_swap8;
 static int music_swap16;
 
 struct _Mix_Music {
-	enum {
-		MUS_CMD,
-		MUS_WAV,
-		MUS_MOD,
-		MUS_MID,
-		MUS_OGG,
-		MUS_MP3
-	} type;
+	Mix_MusicType type;
 	union {
 #ifdef CMD_MUSIC
 		MusicCMD *cmd;
@@ -102,7 +108,7 @@ struct _Mix_Music {
 #ifdef WAV_MUSIC
 		WAVStream *wave;
 #endif
-#ifdef MOD_MUSIC
+#if defined(MOD_MUSIC) || defined(LIBMIKMOD_MUSIC)
 		UNIMOD *module;
 #endif
 #ifdef MID_MUSIC
@@ -121,7 +127,6 @@ struct _Mix_Music {
 #endif
 	} data;
 	Mix_Fading fading;
-	int fade_volume;
 	int fade_step;
 	int fade_steps;
 	int error;
@@ -136,12 +141,20 @@ static int native_midi_ok;
 #endif
 #endif
 
+/* Reference for converting mikmod output to 4/6 channels */
+static int current_output_channels;
+static Uint16 current_output_format;
+
 /* Used to calculate fading steps */
 static int ms_per_step;
 
 /* Local low-level functions prototypes */
-static void lowlevel_halt(void);
-static int  lowlevel_play(Mix_Music *music);
+static void music_internal_initialize_volume(void);
+static void music_internal_volume(int volume);
+static int  music_internal_play(Mix_Music *music, double position);
+static int  music_internal_position(double position);
+static int  music_internal_playing();
+static void music_internal_halt(void);
 
 
 /* Support for hooking when the music has finished */
@@ -158,53 +171,45 @@ void Mix_HookMusicFinished(void (*music_finished)(void))
 /* Mixing function */
 void music_mixer(void *udata, Uint8 *stream, int len)
 {
-	if ( music_playing ) {
-		if ( music_stopped ) {
-			/* To avoid concurrency problems and the use of mutexes,
-			   the music is always stopped from the sound thread */
-			lowlevel_halt(); /* This function sets music_playing to NULL */
-			return;
-		}
+	if ( music_playing && music_active ) {
 		/* Handle fading */
 		if ( music_playing->fading != MIX_NO_FADING ) {
 			if ( music_playing->fade_step++ < music_playing->fade_steps ) {
-				int fade_volume = music_playing->fade_volume;
+				int volume;
 				int fade_step = music_playing->fade_step;
 				int fade_steps = music_playing->fade_steps;
 
 				if ( music_playing->fading == MIX_FADING_OUT ) {
-					Mix_VolumeMusic((fade_volume * (fade_steps-fade_step))
-									/ fade_steps);
+					volume = (music_volume * (fade_steps-fade_step)) / fade_steps;
 				} else { /* Fading in */
-					Mix_VolumeMusic((fade_volume * fade_step) / fade_steps);
+					volume = (music_volume * fade_step) / fade_steps;
 				}
+				music_internal_volume(volume);
 			} else {
 				if ( music_playing->fading == MIX_FADING_OUT ) {
-					lowlevel_halt();
+					music_internal_halt();
+					if ( music_finished_hook ) {
+						music_finished_hook();
+					}
 					return;
 				}
 				music_playing->fading = MIX_NO_FADING;
 			}
 		}
 		/* Restart music if it has to loop */
-		if ( !Mix_PlayingMusic() ) {
-			/* Restart music if it has to loop */
+		if ( !music_internal_playing() ) {
+			/* Restart music if it has to loop at a high level */
 			if ( music_loops && --music_loops ) {
-				Mix_RewindMusic();
-				if ( lowlevel_play(music_playing) < 0 ) {
-					fprintf(stderr,"Warning: Music restart failed.\n");
-					music_stopped = 1; /* Something went wrong */
-					music_playing->fading = MIX_NO_FADING;
+				Mix_Fading current_fade = music_playing->fading;
+				music_internal_play(music_playing, 0.0);
+				music_playing->fading = current_fade;
+			} else {
+				music_internal_halt();
+				if ( music_finished_hook ) {
+					music_finished_hook();
 				}
+				return;
 			}
-			else if (music_finished_hook) {
-			    lowlevel_halt();
-			    music_finished_hook();
-			    return;
-			}
-		}
-		if ( music_volume <= 0 ) { /* Don't mix if volume is null */
-			return;
 		}
 		switch (music_playing->type) {
 #ifdef CMD_MUSIC
@@ -217,9 +222,59 @@ void music_mixer(void *udata, Uint8 *stream, int len)
 				WAVStream_PlaySome(stream, len);
 				break;
 #endif
-#ifdef MOD_MUSIC
+#if defined(MOD_MUSIC) || defined(LIBMIKMOD_MUSIC)
 			case MUS_MOD:
-				VC_WriteBytes((SBYTE *)stream, len);
+				if (current_output_channels > 2) {
+					int small_len = 2 * len / current_output_channels;
+					int i;
+					Uint8 *src, *dst;
+
+					VC_WriteBytes((SBYTE *)stream, small_len);
+					/* and extend to len by copying channels */
+					src = stream + small_len;
+					dst = stream + len;
+
+					switch (current_output_format & 0xFF) {
+						case 8:
+							for ( i=small_len/2; i; --i ) {
+								src -= 2;
+								dst -= current_output_channels;
+								dst[0] = src[0];
+								dst[1] = src[1];
+								dst[2] = src[0];
+								dst[3] = src[1];
+								if (current_output_channels == 6) {
+									dst[4] = src[0];
+									dst[5] = src[1];
+								}
+							}
+							break;
+						case 16:
+							for ( i=small_len/4; i; --i ) {
+								src -= 4;
+								dst -= 2 * current_output_channels;
+								dst[0] = src[0];
+								dst[1] = src[1];
+								dst[2] = src[2];
+								dst[3] = src[3];
+								dst[4] = src[0];
+								dst[5] = src[1];
+								dst[6] = src[2];
+								dst[7] = src[3];
+								if (current_output_channels == 6) {
+									dst[8] = src[0];
+									dst[9] = src[1];
+									dst[10] = src[2];
+									dst[11] = src[3];
+								}
+							}
+							break;
+					}
+
+
+
+				}
+				else VC_WriteBytes((SBYTE *)stream, len);
 				if ( music_swap8 ) {
 					Uint8 *dst;
 					int i;
@@ -281,7 +336,7 @@ int open_music(SDL_AudioSpec *mixer)
 		++music_error;
 	}
 #endif
-#ifdef MOD_MUSIC
+#if defined(MOD_MUSIC) || defined(LIBMIKMOD_MUSIC)
 	/* Set the MikMod music format */
 	music_swap8 = 0;
 	music_swap16 = 0;
@@ -315,8 +370,10 @@ int open_music(SDL_AudioSpec *mixer)
 			++music_error;
 		}
 	}
+	current_output_channels = mixer->channels;
+	current_output_format = mixer->format;
 	if ( mixer->channels > 1 ) {
-		if ( mixer->channels > 2 ) {
+		if ( mixer->channels > MAX_OUTPUT_CHANNELS ) {
 			Mix_SetError("Hardware uses more channels than mixer");
 			++music_error;
 		}
@@ -329,8 +386,17 @@ int open_music(SDL_AudioSpec *mixer)
 	md_sndfxvolume = 128;
 	md_pansep	  = 128;
 	md_reverb	  = 0;
-	MikMod_RegisterAllLoaders();
+#ifdef LIBMIKMOD_MUSIC
+	md_mode |= DMODE_HQMIXER|DMODE_SOFT_MUSIC|DMODE_SURROUND;
+#endif
+#ifdef LIBMIKMOD_MUSIC
+	if(!MikMod_InfoDriver())
+#endif
 	MikMod_RegisterAllDrivers();
+#ifdef LIBMIKMOD_MUSIC
+	if(!MikMod_InfoLoader())
+#endif
+	MikMod_RegisterAllLoaders();
 	if ( MikMod_Init() ) {
 		Mix_SetError("%s", MikMod_strerror(MikMod_errno));
 		++music_error;
@@ -394,7 +460,7 @@ Mix_Music *Mix_LoadMUS(const char *file)
 {
 	FILE *fp;
 	char *ext;
-	Uint8 magic[5];
+	Uint8 magic[5], moremagic[9];
 	Mix_Music *music;
 
 	/* Figure out what kind of file this is */
@@ -406,7 +472,12 @@ Mix_Music *Mix_LoadMUS(const char *file)
 		Mix_SetError("Couldn't read from '%s'", file);
 		return(NULL);
 	}
+	if (!fread(moremagic, 8, 1, fp)) {
+		Mix_SetError("Couldn't read from '%s'", file);
+		return(NULL);
+	}
 	magic[4] = '\0';
+	moremagic[8] = '\0';
 	fclose(fp);
 
 	/* Figure out the file extension, so we can determine the type */
@@ -435,7 +506,7 @@ Mix_Music *Mix_LoadMUS(const char *file)
 	   AIFF files have the magic 12 bytes "FORM" XXXX "AIFF"
 	 */
 	if ( (ext && MIX_string_equals(ext, "WAV")) ||
-	     (strcmp((char *)magic, "RIFF") == 0) ||
+	     ((strcmp((char *)magic, "RIFF") == 0) && (strcmp((char *)(moremagic+4), "WAVE") == 0)) ||
 	     (strcmp((char *)magic, "FORM") == 0) ) {
 		music->type = MUS_WAV;
 		music->data.wave = WAVStream_LoadSong(file, (char *)magic);
@@ -449,7 +520,9 @@ Mix_Music *Mix_LoadMUS(const char *file)
 	/* MIDI files have the magic four bytes "MThd" */
 	if ( (ext && MIX_string_equals(ext, "MID")) ||
 	     (ext && MIX_string_equals(ext, "MIDI")) ||
-	     strcmp((char *)magic, "MThd") == 0 ) {
+	     strcmp((char *)magic, "MThd") == 0  ||
+	     ( strcmp((char *)magic, "RIFF") == 0  &&
+	  	strcmp((char *)(moremagic+4), "RMID") == 0 ) ) {
 		music->type = MUS_MID;
 #ifdef USE_NATIVE_MIDI
   		if ( native_midi_ok ) {
@@ -487,6 +560,7 @@ Mix_Music *Mix_LoadMUS(const char *file)
 #endif
 #ifdef MP3_MUSIC
 	if ( (ext && MIX_string_equals(ext, "MPG")) ||
+	     (ext && MIX_string_equals(ext, "MP3")) ||
 	     (ext && MIX_string_equals(ext, "MPEG")) ||
 	     magic[0]==0xFF && (magic[1]&0xF0)==0xF0) {
 		SMPEG_Info info;
@@ -500,7 +574,7 @@ Mix_Music *Mix_LoadMUS(const char *file)
 		}
 	} else
 #endif
-#ifdef MOD_MUSIC
+#if defined(MOD_MUSIC) || defined(LIBMIKMOD_MUSIC)
 	if ( 1 ) {
 		music->type = MUS_MOD;
 		music->data.module = MikMod_LoadSong((char *)file, 64);
@@ -535,16 +609,20 @@ Mix_Music *Mix_LoadMUS(const char *file)
 void Mix_FreeMusic(Mix_Music *music)
 {
 	if ( music ) {
-		/* Caution: If music is playing, mixer will crash */
-		if ( music == music_playing && !music_stopped ) {
-			if ( music->fading == MIX_FADING_OUT ) {
-				/* Wait for the fade out to finish */
-				while ( music_playing && !music_stopped && (music_playing->fading == MIX_FADING_OUT) )
-					SDL_Delay(100);
-			} else {
-				Mix_HaltMusic(); /* Stop it immediately */
+		/* Stop the music if it's currently playing */
+		SDL_LockAudio();
+		if ( music == music_playing ) {
+			/* Wait for any fade out to finish */
+			while ( music->fading == MIX_FADING_OUT ) {
+				SDL_UnlockAudio();
+				SDL_Delay(100);
+				SDL_LockAudio();
+			}
+			if ( music == music_playing ) {
+				music_internal_halt();
 			}
 		}
+		SDL_UnlockAudio();
 		switch (music->type) {
 #ifdef CMD_MUSIC
 			case MUS_CMD:
@@ -556,7 +634,7 @@ void Mix_FreeMusic(Mix_Music *music)
 				WAVStream_FreeSong(music->data.wave);
 				break;
 #endif
-#ifdef MOD_MUSIC
+#if defined(MOD_MUSIC) || defined(LIBMIKMOD_MUSIC)
 			case MUS_MOD:
 				MikMod_FreeSong(music->data.module);
 				break;
@@ -593,190 +671,305 @@ void Mix_FreeMusic(Mix_Music *music)
 	}
 }
 
-static int lowlevel_play(Mix_Music *music)
+/* Find out the music format of a mixer music, or the currently playing
+   music, if 'music' is NULL.
+*/
+Mix_MusicType Mix_GetMusicType(const Mix_Music *music)
 {
-	if(!music)
-		return(-1);
+	Mix_MusicType type = MUS_NONE;
 
-	switch (music->type) {
-#ifdef CMD_MUSIC
-		case MUS_CMD:
-			MusicCMD_SetVolume(music_volume);
-			MusicCMD_Start(music->data.cmd);
-			break;
-#endif
-#ifdef WAV_MUSIC
-		case MUS_WAV:
-			WAVStream_SetVolume(music_volume);
-			WAVStream_Start(music->data.wave);
-			break;
-#endif
-#ifdef MOD_MUSIC
-		case MUS_MOD:
-			Player_SetVolume((SWORD)music_volume);
-			Player_Start(music->data.module);
-			Player_SetPosition(0);
-			break;
-#endif
-#ifdef MID_MUSIC
-		case MUS_MID:
-#ifdef USE_NATIVE_MIDI
-  			if ( native_midi_ok ) {
-				native_midi_setvolume(music_volume);
-				native_midi_start(music->data.nativemidi);
-			} MIDI_ELSE
-#endif
-#ifdef USE_TIMIDITY_MIDI
-			if ( timidity_ok ) {
-				Timidity_SetVolume(music_volume);
-				Timidity_Start(music->data.midi);
-			}
-#endif
-			break;
-#endif
-#ifdef OGG_MUSIC
-		case MUS_OGG:
-			OGG_setvolume(music->data.ogg, music_volume);
-			OGG_play(music->data.ogg);
-			break;
-#endif
-#ifdef MP3_MUSIC
-		case MUS_MP3:
-			SMPEG_enableaudio(music->data.mp3,1);
-			SMPEG_enablevideo(music->data.mp3,0);
-			SMPEG_setvolume(music->data.mp3,(int)(((float)music_volume/(float)MIX_MAX_VOLUME)*100.0));
-			SMPEG_play(music->data.mp3);
-			break;
-#endif
-		default:
-			/* Unknown music type?? */
-			return(-1);
+	if ( music ) {
+		type = music->type;
+	} else {
+		SDL_LockAudio();
+		if ( music_playing ) {
+			type = music_playing->type;
+		}
+		SDL_UnlockAudio();
 	}
-	return(0);
+	return(type);
 }
 
 /* Play a music chunk.  Returns 0, or -1 if there was an error.
-*/
-int Mix_PlayMusic(Mix_Music *music, int loops)
+ */
+static int music_internal_play(Mix_Music *music, double position)
 {
+	int retval = 0;
+
+	/* Note the music we're playing */
+	if ( music_playing ) {
+		music_internal_halt();
+	}
+	music_playing = music;
+
+	/* Set the initial volume */
+	if ( music->type != MUS_MOD ) {
+		music_internal_initialize_volume();
+	}
+
+	/* Set up for playback */
+	switch (music->type) {
+#ifdef CMD_MUSIC
+	    case MUS_CMD:
+		MusicCMD_Start(music->data.cmd);
+		break;
+#endif
+#ifdef WAV_MUSIC
+	    case MUS_WAV:
+		WAVStream_Start(music->data.wave);
+		break;
+#endif
+#if defined(MOD_MUSIC) || defined(LIBMIKMOD_MUSIC)
+	    case MUS_MOD:
+		Player_Start(music->data.module);
+		/* Player_SetVolume() does nothing before Player_Start() */
+		music_internal_initialize_volume();
+		break;
+#endif
+#ifdef MID_MUSIC
+	    case MUS_MID:
+#ifdef USE_NATIVE_MIDI
+		if ( native_midi_ok ) {
+			native_midi_start(music->data.nativemidi);
+		} MIDI_ELSE
+#endif
+#ifdef USE_TIMIDITY_MIDI
+		if ( timidity_ok ) {
+			Timidity_Start(music->data.midi);
+		}
+#endif
+		break;
+#endif
+#ifdef OGG_MUSIC
+	    case MUS_OGG:
+		OGG_play(music->data.ogg);
+		break;
+#endif
+#ifdef MP3_MUSIC
+	    case MUS_MP3:
+		SMPEG_enableaudio(music->data.mp3,1);
+		SMPEG_enablevideo(music->data.mp3,0);
+		SMPEG_play(music_playing->data.mp3);
+		break;
+#endif
+	    default:
+		Mix_SetError("Can't play unknown music type");
+		retval = -1;
+		break;
+	}
+
+	/* Set the playback position, note any errors if an offset is used */
+	if ( retval == 0 ) {
+		if ( position > 0.0 ) {
+			if ( music_internal_position(position) < 0 ) {
+				Mix_SetError("Position not implemented for music type");
+				retval = -1;
+			}
+		} else {
+			music_internal_position(0.0);
+		}
+	}
+
+	/* If the setup failed, we're not playing any music anymore */
+	if ( retval < 0 ) {
+		music_playing = NULL;
+	}
+	return(retval);
+}
+int Mix_FadeInMusicPos(Mix_Music *music, int loops, int ms, double position)
+{
+	int retval;
+
 	/* Don't play null pointers :-) */
 	if ( music == NULL ) {
+		Mix_SetError("music parameter was NULL");
 		return(-1);
 	}
-	/* If the current music is fading out, wait for the fade to complete */
-	while ( music_playing && !music_stopped && music_playing->fading==MIX_FADING_OUT ) {
-		SDL_Delay(100);
-	}
 
-	if ( lowlevel_play(music) < 0 ) {
-		return(-1);
+	/* Setup the data */
+	if ( ms ) {
+		music->fading = MIX_FADING_IN;
+	} else {
+		music->fading = MIX_NO_FADING;
+	}
+	music->fade_step = 0;
+	music->fade_steps = ms/ms_per_step;
+
+	/* Play the puppy */
+	SDL_LockAudio();
+	/* If the current music is fading out, wait for the fade to complete */
+	while ( music_playing && (music_playing->fading == MIX_FADING_OUT) ) {
+		SDL_UnlockAudio();
+		SDL_Delay(100);
+		SDL_LockAudio();
 	}
 	music_active = 1;
-	music_stopped = 0;
 	music_loops = loops;
-	music_playing = music;
-	music_playing->fading = MIX_NO_FADING;
-	return(0);
-}
+	retval = music_internal_play(music, position);
+	SDL_UnlockAudio();
 
-/* Fade in a music over "ms" milliseconds */
+	return(retval);
+}
 int Mix_FadeInMusic(Mix_Music *music, int loops, int ms)
 {
-	if ( music && music_volume > 0 ) { /* No need to fade if we can't hear it */
-		music->fade_volume = music_volume;
-		music_volume = 0;
-		if ( Mix_PlayMusic(music, loops) < 0 ) {
-			return(-1);
+	return Mix_FadeInMusicPos(music, loops, ms, 0.0);
+}
+int Mix_PlayMusic(Mix_Music *music, int loops)
+{
+	return Mix_FadeInMusicPos(music, loops, 0, 0.0);
+}
+
+/* Set the playing music position */
+int music_internal_position(double position)
+{
+	int retval = 0;
+
+	switch (music_playing->type) {
+#if defined(MOD_MUSIC) || defined(LIBMIKMOD_MUSIC)
+	    case MUS_MOD:
+		Player_SetPosition((UWORD)position);
+		break;
+#endif
+#ifdef OGG_MUSIC
+	    case MUS_OGG:
+		OGG_jump_to_time(music_playing->data.ogg, position);
+		break;
+#endif
+#ifdef MP3_MUSIC
+	    case MUS_MP3:
+		if ( position > 0.0 ) {
+			SMPEG_skip(music_playing->data.mp3, position);
+		} else {
+			SMPEG_rewind(music_playing->data.mp3);
+			SMPEG_play(music_playing->data.mp3);
 		}
-		music_playing->fade_step = 0;
-		music_playing->fade_steps = ms/ms_per_step;
-		music_playing->fading = MIX_FADING_IN;
+		break;
+#endif
+	    default:
+		/* TODO: Implement this for other music backends */
+		retval = -1;
+		break;
+	}
+	return(retval);
+}
+int Mix_SetMusicPosition(double position)
+{
+	int retval;
+
+	SDL_LockAudio();
+	if ( music_playing ) {
+		retval = music_internal_position(position);
+		if ( retval < 0 ) {
+			Mix_SetError("Position not implemented for music type");
+		}
+	} else {
+		Mix_SetError("Music isn't playing");
+		retval = -1;
+	}
+	SDL_UnlockAudio();
+
+	return(retval);
+}
+
+/* Set the music's initial volume */
+static void music_internal_initialize_volume(void)
+{
+	if ( music_playing->fading == MIX_FADING_IN ) {
+		music_internal_volume(0);
+	} else {
+		music_internal_volume(music_volume);
 	}
-	return(0);
 }
 
 /* Set the music volume */
+static void music_internal_volume(int volume)
+{
+	switch (music_playing->type) {
+#ifdef CMD_MUSIC
+	    case MUS_CMD:
+		MusicCMD_SetVolume(volume);
+		break;
+#endif
+#ifdef WAV_MUSIC
+	    case MUS_WAV:
+		WAVStream_SetVolume(volume);
+		break;
+#endif
+#if defined(MOD_MUSIC) || defined(LIBMIKMOD_MUSIC)
+	    case MUS_MOD:
+		Player_SetVolume((SWORD)volume);
+		break;
+#endif
+#ifdef MID_MUSIC
+	    case MUS_MID:
+#ifdef USE_NATIVE_MIDI
+		if ( native_midi_ok ) {
+			native_midi_setvolume(volume);
+		} MIDI_ELSE
+#endif
+#ifdef USE_TIMIDITY_MIDI
+		if ( timidity_ok ) {
+			Timidity_SetVolume(volume);
+		}
+#endif
+		break;
+#endif
+#ifdef OGG_MUSIC
+	    case MUS_OGG:
+		OGG_setvolume(music_playing->data.ogg, volume);
+		break;
+#endif
+#ifdef MP3_MUSIC
+	    case MUS_MP3:
+		SMPEG_setvolume(music_playing->data.mp3,(int)(((float)volume/(float)MIX_MAX_VOLUME)*100.0));
+		break;
+#endif
+	    default:
+		/* Unknown music type?? */
+		break;
+	}
+}
 int Mix_VolumeMusic(int volume)
 {
 	int prev_volume;
 
 	prev_volume = music_volume;
 	if ( volume < 0 ) {
-		volume = 0;
+		return prev_volume;
 	}
 	if ( volume > SDL_MIX_MAXVOLUME ) {
 		volume = SDL_MIX_MAXVOLUME;
 	}
 	music_volume = volume;
-	if ( music_playing && !music_stopped ) {
-		switch (music_playing->type) {
-#ifdef CMD_MUSIC
-		case MUS_CMD:
-			MusicCMD_SetVolume(music_volume);
-			break;
-#endif
-#ifdef WAV_MUSIC
-		case MUS_WAV:
-			WAVStream_SetVolume(music_volume);
-			break;
-#endif
-#ifdef MOD_MUSIC
-		case MUS_MOD:
-			Player_SetVolume((SWORD)music_volume);
-			break;
-#endif
-#ifdef MID_MUSIC
-		case MUS_MID:
-#ifdef USE_NATIVE_MIDI
-			if ( native_midi_ok ) {
-				native_midi_setvolume(music_volume);
-			} MIDI_ELSE
-#endif
-#ifdef USE_TIMIDITY_MIDI
-			if ( timidity_ok ) {
-				Timidity_SetVolume(music_volume);
-			}
-#endif
-			break;
-#endif
-#ifdef OGG_MUSIC
-		case MUS_OGG:
-			OGG_setvolume(music_playing->data.ogg, music_volume);
-			break;
-#endif
-#ifdef MP3_MUSIC
-		case MUS_MP3:
-			SMPEG_setvolume(music_playing->data.mp3,(int)(((float)music_volume/(float)MIX_MAX_VOLUME)*100.0));
-			break;
-#endif
-		default:
-			/* Unknown music type?? */
-			break;
-		}
+	SDL_LockAudio();
+	if ( music_playing ) {
+		music_internal_volume(music_volume);
 	}
+	SDL_UnlockAudio();
 	return(prev_volume);
 }
 
-static void lowlevel_halt(void)
+/* Halt playing of music */
+static void music_internal_halt(void)
 {
 	switch (music_playing->type) {
 #ifdef CMD_MUSIC
-	case MUS_CMD:
+	    case MUS_CMD:
 		MusicCMD_Stop(music_playing->data.cmd);
 		break;
 #endif
 #ifdef WAV_MUSIC
-	case MUS_WAV:
+	    case MUS_WAV:
 		WAVStream_Stop();
 		break;
 #endif
-#ifdef MOD_MUSIC
-	case MUS_MOD:
+#if defined(MOD_MUSIC) || defined(LIBMIKMOD_MUSIC)
+	    case MUS_MOD:
 		Player_Stop();
 		break;
 #endif
 #ifdef MID_MUSIC
-	case MUS_MID:
+	    case MUS_MID:
 #ifdef USE_NATIVE_MIDI
 		if ( native_midi_ok ) {
 			native_midi_stop();
@@ -790,108 +983,77 @@ static void lowlevel_halt(void)
 		break;
 #endif
 #ifdef OGG_MUSIC
-	case MUS_OGG:
+	    case MUS_OGG:
 		OGG_stop(music_playing->data.ogg);
 		break;
 #endif
 #ifdef MP3_MUSIC
-	case MUS_MP3:
+	    case MUS_MP3:
 		SMPEG_stop(music_playing->data.mp3);
 		break;
 #endif
-	default:
+	    default:
 		/* Unknown music type?? */
 		return;
 	}
-	if ( music_playing->fading != MIX_NO_FADING ) /* Restore volume */
-		music_volume = music_playing->fade_volume;
 	music_playing->fading = MIX_NO_FADING;
 	music_playing = NULL;
-	music_active = 0;
-	music_loops = 0;
-	music_stopped = 0;
 }
-
-/* Halt playing of music */
 int Mix_HaltMusic(void)
 {
-	if ( music_playing && !music_stopped ) {
-		/* Mark the music to be stopped from the sound thread */
-		music_stopped = 1;
-		/* Wait for it to be actually stopped */
-		while ( music_playing && music_active )
-			SDL_Delay(10);
+	SDL_LockAudio();
+	if ( music_playing ) {
+		music_internal_halt();
 	}
+	SDL_UnlockAudio();
+
 	return(0);
 }
 
 /* Progressively stop the music */
 int Mix_FadeOutMusic(int ms)
 {
-	if ( music_playing && !music_stopped &&
-	     (music_playing->fading == MIX_NO_FADING) ) {
-		if ( music_volume > 0 ) {
-			music_playing->fading = MIX_FADING_OUT;
-			music_playing->fade_volume = music_volume;
-			music_playing->fade_step = 0;
-			music_playing->fade_steps = ms/ms_per_step;
-			return(1);
-		}
+	int retval = 0;
+
+	SDL_LockAudio();
+	if ( music_playing && (music_playing->fading == MIX_NO_FADING) ) {
+		music_playing->fading = MIX_FADING_OUT;
+		music_playing->fade_step = 0;
+		music_playing->fade_steps = ms/ms_per_step;
+		retval = 1;
 	}
-	return(0);
+	SDL_UnlockAudio();
+
+	return(retval);
 }
 
 Mix_Fading Mix_FadingMusic(void)
 {
-	if( music_playing && !music_stopped )
-		return music_playing->fading;
-	return MIX_NO_FADING;
+	Mix_Fading fading = MIX_NO_FADING;
+
+	SDL_LockAudio();
+	if ( music_playing ) {
+		fading = music_playing->fading;
+	}
+	SDL_UnlockAudio();
+
+	return(fading);
 }
 
 /* Pause/Resume the music stream */
 void Mix_PauseMusic(void)
 {
-	if ( music_playing && !music_stopped ) {
-		music_active = 0;
-	}
+	music_active = 0;
 }
 
 void Mix_ResumeMusic(void)
 {
-	if ( music_playing && !music_stopped ) {
-		music_active = 1;
-	}
+	music_active = 1;
 }
 
 void Mix_RewindMusic(void)
 {
-	if ( music_playing && !music_stopped ) {
-		switch ( music_playing->type ) {
-#ifdef MOD_MUSIC
-		case MUS_MOD:
-			Player_Start(music_playing->data.module);
-			Player_SetPosition(0);
-			break;
-#endif
-#ifdef MP3_MUSIC
-		case MUS_MP3:
-			SMPEG_rewind(music_playing->data.mp3);
-			break;
-#endif
-#ifdef MID_MUSIC
-		case MUS_MID:
-#ifdef USE_NATIVE_MIDI
-			if ( native_midi_ok ) {
-				native_midi_stop();
-			}
-#endif
-			break;
-#endif
-		default:
-			/* TODO: Implement this for other music backends */
-			break;
-		}
-	}
+	Mix_SetMusicPosition(0.0);
 }
 
 int Mix_PausedMusic(void)
@@ -900,66 +1062,77 @@ int Mix_PausedMusic(void)
 }
 
 /* Check the status of the music */
-int Mix_PlayingMusic(void)
+static int music_internal_playing()
 {
-	if ( music_playing && ! music_stopped ) {
-		switch (music_playing->type) {
+	int playing = 1;
+	switch (music_playing->type) {
 #ifdef CMD_MUSIC
-			case MUS_CMD:
-				if (!MusicCMD_Active(music_playing->data.cmd)) {
-					return(0);
-				}
-				break;
+	    case MUS_CMD:
+		if (!MusicCMD_Active(music_playing->data.cmd)) {
+			playing = 0;
+		}
+		break;
 #endif
 #ifdef WAV_MUSIC
-			case MUS_WAV:
-				if ( ! WAVStream_Active() ) {
-					return(0);
-				}
-				break;
+	    case MUS_WAV:
+		if ( ! WAVStream_Active() ) {
+			playing = 0;
+		}
+		break;
 #endif
-#ifdef MOD_MUSIC
-			case MUS_MOD:
-				if ( ! Player_Active() ) {
-					return(0);
-				}
-				break;
+#if defined(MOD_MUSIC) || defined(LIBMIKMOD_MUSIC)
+	    case MUS_MOD:
+		if ( ! Player_Active() ) {
+			playing = 0;
+		}
+		break;
 #endif
 #ifdef MID_MUSIC
-			case MUS_MID:
+	    case MUS_MID:
 #ifdef USE_NATIVE_MIDI
-				if ( native_midi_ok ) {
-					if ( ! native_midi_active() )
-						return(0);
-				} MIDI_ELSE
+		if ( native_midi_ok ) {
+			if ( ! native_midi_active() )
+				playing = 0;
+		} MIDI_ELSE
 #endif
 #ifdef USE_TIMIDITY_MIDI
-				if ( timidity_ok ) {
-					if ( ! Timidity_Active() )
-						return(0);
-				}
+		if ( timidity_ok ) {
+			if ( ! Timidity_Active() )
+				playing = 0;
+		}
 #endif
-				break;
+		break;
 #endif
 #ifdef OGG_MUSIC
-			case MUS_OGG:
-				if ( ! OGG_playing(music_playing->data.ogg) ) {
-					return(0);
-				}
-				break;
+	    case MUS_OGG:
+		if ( ! OGG_playing(music_playing->data.ogg) ) {
+			playing = 0;
+		}
+		break;
 #endif
 #ifdef MP3_MUSIC
-			case MUS_MP3:
-				if ( SMPEG_status(music_playing->data.mp3) != SMPEG_PLAYING )
-					return(0);
-				break;
+	    case MUS_MP3:
+		if ( SMPEG_status(music_playing->data.mp3) != SMPEG_PLAYING )
+			playing = 0;
+		break;
 #endif
-			default:
-				break;
-		}
-		return(1);
+	    default:
+		playing = 0;
+		break;
 	}
-	return(0);
+	return(playing);
+}
+int Mix_PlayingMusic(void)
+{
+	int playing = 0;
+
+	SDL_LockAudio();
+	if ( music_playing ) {
+		playing = music_internal_playing();
+	}
+	SDL_UnlockAudio();
+
+	return(playing);
 }
 
 /* Set the external music playback command */
@@ -980,6 +1153,65 @@ int Mix_SetMusicCMD(const char *command)
 	return(0);
 }
 
+#ifdef LIBMIKMOD_MUSIC
+static int _pl_synchro_value;
+void Player_SetSynchroValue(int i)
+{
+	fprintf(stderr,"SDL_mixer: Player_SetSynchroValue is not supported.\n");
+	_pl_synchro_value = i;
+}
+
+int Player_GetSynchroValue(void)
+{
+	fprintf(stderr,"SDL_mixer: Player_GetSynchroValue is not supported.\n");
+	return _pl_synchro_value;
+}
+#endif
+
+int Mix_SetSynchroValue(int i)
+{
+	if ( music_playing && ! music_stopped ) {
+		switch (music_playing->type) {
+#if defined(MOD_MUSIC) || defined(LIBMIKMOD_MUSIC)
+		    case MUS_MOD:
+			if ( ! Player_Active() ) {
+				return(-1);
+			}
+			Player_SetSynchroValue(i);
+			return 0;
+			break;
+#endif
+		    default:
+			return(-1);
+			break;
+		}
+		return(-1);
+	}
+	return(-1);
+}
+
+int Mix_GetSynchroValue(void)
+{
+	if ( music_playing && ! music_stopped ) {
+		switch (music_playing->type) {
+#if defined(MOD_MUSIC) || defined(LIBMIKMOD_MUSIC)
+		    case MUS_MOD:
+			if ( ! Player_Active() ) {
+				return(-1);
+			}
+			return Player_GetSynchroValue();
+			break;
+#endif
+		    default:
+			return(-1);
+			break;
+		}
+		return(-1);
+	}
+	return(-1);
+}
+
+
 /* Uninitialize the music players */
 void close_music(void)
 {
@@ -987,10 +1219,140 @@ void close_music(void)
 #ifdef CMD_MUSIC
 	Mix_SetMusicCMD(NULL);
 #endif
-#ifdef MOD_MUSIC
+#if defined(MOD_MUSIC) || defined(LIBMIKMOD_MUSIC)
 	MikMod_Exit();
+# ifndef LIBMIKMOD_MUSIC
 	MikMod_UnregisterAllLoaders();
 	MikMod_UnregisterAllDrivers();
+# endif
 #endif
 }
 
+#ifdef USE_RWOPS
+
+# ifdef LIBMIKMOD_MUSIC
+typedef struct
+{
+	MREADER mr;
+	int offset;
+	int eof;
+	SDL_RWops *rw;
+} LMM_MREADER;
+BOOL LMM_Seek(struct MREADER *mr,long to,int dir)
+{
+	int at;
+	LMM_MREADER* lmmmr=(LMM_MREADER*)mr;
+	if(dir==SEEK_SET)
+		to+=lmmmr->offset;
+	at=SDL_RWseek(lmmmr->rw, to, dir);
+	return at<lmmmr->offset;
+}
+long LMM_Tell(struct MREADER *mr)
+{
+	int at;
+	LMM_MREADER* lmmmr=(LMM_MREADER*)mr;
+	at=SDL_RWtell(lmmmr->rw)-lmmmr->offset;
+	return at;
+}
+BOOL LMM_Read(struct MREADER *mr,void *buf,size_t sz)
+{
+	int got;
+	LMM_MREADER* lmmmr=(LMM_MREADER*)mr;
+	got=SDL_RWread(lmmmr->rw, buf, sz, 1);
+	return got;
+}
+int LMM_Get(struct MREADER *mr)
+{
+	unsigned char c;
+	int i=EOF;
+	LMM_MREADER* lmmmr=(LMM_MREADER*)mr;
+	if(SDL_RWread(lmmmr->rw,&c,1,1))
+		i=c;
+	return i;
+}
+BOOL LMM_Eof(struct MREADER *mr)
+{
+	int offset;
+	LMM_MREADER* lmmmr=(LMM_MREADER*)mr;
+	offset=LMM_Tell(mr);
+	return offset>=lmmmr->eof;
+}
+MODULE *MikMod_LoadSongRW(SDL_RWops *rw, int maxchan)
+{
+	LMM_MREADER lmmmr={
+		LMM_Seek,
+		LMM_Tell,
+		LMM_Read,
+		LMM_Get,
+		LMM_Eof,
+		0,
+		0,
+		rw
+	};
+	MODULE *m;
+	lmmmr.offset=SDL_RWtell(rw);
+	SDL_RWseek(rw,0,SEEK_END);
+	lmmmr.eof=SDL_RWtell(rw);
+	SDL_RWseek(rw,lmmmr.offset,SEEK_SET);
+	m=Player_LoadGeneric((MREADER*)&lmmmr,maxchan,0);
+	return m;
+}
+# endif
+
+Mix_Music *Mix_LoadMUS_RW(SDL_RWops *rw) {
+	Uint8 magic[5];	  /*Apparently there is no way to check if the file is really a MOD,*/
+	/*		    or there are too many formats supported by MikMod or MikMod does */
+	/*		    this check by itself. If someone implements other formats (e.g. MP3) */
+	/*		    the check can be uncommented */
+	Mix_Music *music;
+	int start;
+
+	/* Figure out what kind of file this is */
+	start = SDL_RWtell(rw);
+	if (SDL_RWread(rw,magic,1,4)!=4) {
+		Mix_SetError("Couldn't read from RWops");
+		return NULL;
+	}
+	SDL_RWseek(rw, start, SEEK_SET);
+	magic[4]='\0';
+
+	/* Allocate memory for the music structure */
+	music=(Mix_Music *)malloc(sizeof(Mix_Music));
+	if (music==NULL ) {
+		Mix_SetError("Out of memory");
+		return(NULL);
+	}
+	music->error = 0;
+
+#ifdef OGG_MUSIC
+	/* Ogg Vorbis files have the magic four bytes "OggS" */
+	if ( strcmp((char *)magic, "OggS") == 0 ) {
+		music->type = MUS_OGG;
+		music->data.ogg = OGG_new_RW(rw);
+		if ( music->data.ogg == NULL ) {
+			music->error = 1;
+		}
+	} else
+#endif
+#if defined(MOD_MUSIC) || defined(LIBMIKMOD_MUSIC)
+	if (1) {
+		music->type=MUS_MOD;
+		music->data.module=MikMod_LoadSongRW(rw,64);
+		if (music->data.module==NULL) {
+			Mix_SetError("%s",MikMod_strerror(MikMod_errno));
+			music->error=1;
+		}
+	} else
+#endif
+	{
+		Mix_SetError("Unrecognized music format");
+		music->error=1;
+	}
+	if (music->error) {
+		free(music);
+		music=NULL;
+	}
+	return(music);
+}
+
+#endif /* USE_RWOPS */
